/*
 * Copyright (c) 2002 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: winfmpcmd.c,v 1.1.1.1 2002/07/30 01:53:08 zinnia Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#include <fcntl.h>
#include "winfmp_constants.h"
#include "winfmpcmd.h"

static char* rcsid = "$Id: winfmpcmd.c,v 1.1.1.1 2002/07/30 01:53:08 zinnia Exp $";

version winfmp_dll;
version winfmp_interface;
version myself = {MYSELF_VERSION_MAJOR, MYSELF_VERSION_MINOR};

static void version_set(int d, version* p)
{
  (*p).major = d / 100;
  (*p).minor = d % 100;
}

void usage()
{
  fprintf(stderr, "winfmpcmd version %d.%d  (C) Zinnia\n%s\nusage: winfmpcmd [options] [filenames...]\n"
	  "\t-r [rate]: set rate to [rate]\n"
	  "\t-l [loop]: set loopcount to [loop] (0 = infinite)\n"
	  "\t-5: calc FM generator in 55K(interally)\n"
	  "\t-p [path]: set rhythm path\n",
	  myself.major,
	  myself.minor,
	  rcsid);
  exit(1);
}

#define SAMPLES 256
int play_loop(char* filename, int loop)
{
  int load_result;
  int curloop;
  static char buf[SAMPLES * 2 * sizeof(short)];

  load_result = fmp_load(filename);
  fmp_start();
  if(load_result != WINFMP_OK){
    fprintf(stderr, "%s: fmp_load: ", filename);
    switch(load_result){
    case FMP_ERR_OPEN_MUSIC_FILE:
      fprintf(stderr, "can't open music file.\n");
      return -1;
      break; /* NOTREACHED */
    case FMP_ERR_WRONG_MUSIC_FILE:
      fprintf(stderr, "wrong music file (not an fmp file?).\n");
      return -1;
      break; /* NOTREACHED */
   case FMP_ERR_OPEN_PVI_FILE:
      fprintf(stderr, "can't open pvi file.\n");
      return -1;
      break; /* NOTREACHED */
    case FMP_ERR_OPEN_PPZ1_FILE:
      fprintf(stderr, "can't open ppz1 file.\n");
      return -1;
      break; /* NOTREACHED */
    case FMP_ERR_WRONG_PVI_FILE:
      fprintf(stderr, "wrong pvi file.\n");
      return -1;
      break; /* NOTREACHED */
    case FMP_ERR_WRONG_PPZ1_FILE:
      fprintf(stderr, "wrong ppz1 file.\n");
      return -1;
      break; /* NOTREACHED */
    case FMP_WARNING_PVI_ALREADY_LOAD:
      fprintf(stderr, "(warning) pvi already loaded.\n");
      break;
    case FMP_WARNING_PPZ1_ALREADY_LOAD:
      fprintf(stderr, "(warning) ppz1 already loaded.\n");
      break;
    case FMP_ERR_OUT_OF_MEMORY:
      fprintf(stderr, "out of memory.\n");
      exit(1);
      break; /* NOTREACHED */
    }
  }

  fprintf(stderr, "%s: start playing (loop %d)...\n", filename, loop);
  memset(buf, 0, SAMPLES * sizeof(short) * 2);
  while(1){
    curloop = fmp_getloopcount();
    if(curloop == -1) break;
    if(loop > 0 && curloop >= loop){
      fmp_fadeout(0);
    }
    fmp_getpcmdata((short*)buf, SAMPLES);
    write(1, buf, sizeof(short) * 2 * SAMPLES);
  }
  fmp_stop();
  return 0;
}

typedef struct {
  int rate;
  int loop;
  char* rhythmpath;
  BOOL in55k;
} opt_context;

int play_files(int curarg, int argc, char* argv[], opt_context* opt)
{
  HMODULE h;
  int i;
  int played = 0;
  int failed = 0;
  int total = 0;

  h = LoadLibrary("WinFMP.DLL");
  if(h == NULL){
    fprintf(stderr, "LoadLibrary failed.\n");
    return 0;
  }

#define LOAD_WRAPPER(signature) if(LOAD_FUNCTION(h,signature) == NULL) { fprintf(stderr, "GetProcAddress(" #signature ") failed.\n"); exit(1); }

  LOAD_WRAPPER(fmp_getversion);
  LOAD_WRAPPER(fmp_getinterfaceversion);
  LOAD_WRAPPER(fmp_init);
  LOAD_WRAPPER(fmp_setpcmrate);
  LOAD_WRAPPER(fmp_load);
  LOAD_WRAPPER(fmp_start);
  LOAD_WRAPPER(fmp_getpcmdata);
  LOAD_WRAPPER(fmp_stop);
  LOAD_WRAPPER(fmp_setfmcalc55k);
  LOAD_WRAPPER(fmp_getloopcount);
  LOAD_WRAPPER(fmp_setpcmdir);
  LOAD_WRAPPER(fmp_fadeout);
  version_set(fmp_getversion(), &winfmp_dll);
  version_set(fmp_getinterfaceversion(), &winfmp_interface);

  fprintf(stderr, "WinFMP.DLL version %d.%d, interface version %d.%d\n", 
	 winfmp_dll.major, winfmp_dll.minor,
	 winfmp_interface.major, winfmp_interface.minor);

  if(fmp_init(opt->rhythmpath) == FALSE){
    fprintf(stderr, "fmp_init: failed.\n");
    return 0;
  }

  fmp_setpcmrate(opt->rate);
  fmp_setfmcalc55k(opt->in55k);

  setmode(1, _O_BINARY);

  for(i = curarg; i < argc; i++){
    switch(play_loop(argv[i], opt->loop)){
    case -1:
      failed++;
      break;
    case 0:
      played++;
      break;
    }
    total++;
  }

  if(total == 0){
    fprintf(stderr, "no files to play.\n");
  } else {
    fprintf(stderr, "%d files played. %d files failed.\n", played, failed);
  }
  return total;
}

int main(int argc, char* argv[])
{
  opt_context opt;
  char* arg;
  int i;
  int total = 0;

  opt.rhythmpath = DEFAULT_RHYTHMPATH;
  opt.rate = DEFAULT_RATE;
  opt.loop = DEFAULT_LOOP;
  opt.in55k = FALSE;

  {
    char* tmp = getenv("FMGEN_RHYTHMPATH");
    if(tmp != NULL) opt.rhythmpath = tmp;
  }

  for(i = 1; i < argc; i++){
    #define NEXT_ARG() (*(argv[i] + 2) == '\0' ? argv[++i] : argv[i] + 2)
    if(*argv[i] == '-'){
      switch(*(argv[i] + 1)){
      case 'r':
	arg = NEXT_ARG();
	opt.rate = atoi(arg);
	break;
      case 'l':
	arg = NEXT_ARG();
	opt.loop = atoi(arg);
	break;
      case '5':
	opt.in55k = TRUE;
	break;
      case 'p':
	arg = NEXT_ARG();
	opt.rhythmpath = arg;
      case 'h':
	/* FALLTHROUGH */
      default:
	usage();
	break; /* NOTREACHED */
      }
    } else {
      /* filename */
      total = play_files(i, argc, argv, &opt);
      break;
    }
  }
  if(total == 0) usage();

  return 0;
}
